@@ -6,6 +6,13 @@ if six.PY2:
 else:
     import json
 
+# Use cPickle in python 2 and pickle in python 3
+try:
+    import cPickle as pickle
+except ImportError:
+    import pickle
+
+# Adapt hashlib.md5 to eat str in python 3
 if six.PY2:
     md5 = hashlib.md5
 else:
@@ -20,3 +27,6 @@ else:
 
         def hexdigest(self):
             return self.md5.hexdigest()
+
+def md5hex(s):
+    return md5(s).hexdigest()
